use super::{
    expr::{Expr, ExprOp},
    opcode::OpCode,
    value::Value,
    Addr, Error, Register,
};
use std::collections::HashMap;

pub struct Compiler<'a> {
    expr: &'a Expr,
    current_register: Register,
    labels: Addr,
}

mod optimizations;

impl<'a> Compiler<'a> {
    pub fn new(expr: &'a Expr) -> Self {
        Self {
            expr,
            current_register: 0.into(),
            labels: 0.into(),
        }
    }

    fn next_label(&mut self) -> Addr {
        *self.labels += 1;
        self.labels
    }

    fn next_register(&mut self) -> Register {
        *self.current_register += 1;
        self.current_register
    }

    fn compile_expr(
        &mut self,
        expr: &'a Expr,
        exit_label: Addr,
    ) -> Result<(Vec<OpCode>, Register), Error> {
        let mut return_value = self.next_register();
        Ok(match expr {
            Expr::Variable(name) => (
                vec![OpCode::LOAD_EXTERNAL(return_value, name.clone())],
                return_value,
            ),
            Expr::Bool(value) => (
                vec![OpCode::LOAD(return_value, Value::Bool(*value))],
                return_value,
            ),
            Expr::String(string) => (
                vec![OpCode::LOAD(return_value, Value::String(string.clone()))],
                return_value,
            ),
            Expr::Number(number) => (
                vec![OpCode::LOAD(return_value, Value::Number(*number))],
                return_value,
            ),
            Expr::Op { op, terms } => {
                let mut opcodes = vec![];
                let current_exit_label = self.next_label();

                let mut compiled_terms = terms
                    .iter()
                    .map(|term| self.compile_expr(term, current_exit_label))
                    .collect::<Result<Vec<_>, _>>()?;

                match op {
                    ExprOp::Not => {
                        let (mut opcodes, term_return_value) =
                            compiled_terms.pop().ok_or(Error::UnexpectedExprState)?;
                        if !compiled_terms.is_empty() {
                            return Err(Error::UnexpectedExprState);
                        }
                        opcodes.push(OpCode::NOT(return_value, return_value));
                        return_value = term_return_value;
                    }
                    ExprOp::Eq => {
                        let mut iter = compiled_terms.into_iter();
                        let (mut term_opcodes, last_value) =
                            iter.next().ok_or(Error::MissingNextRule)?;

                        opcodes.append(&mut term_opcodes);
                        if iter
                            .map(|(mut term_codes, term_return)| {
                                opcodes.append(&mut term_codes);
                                opcodes.push(OpCode::EQ(return_value, last_value, term_return));
                                opcodes.push(OpCode::JNE(return_value, exit_label));
                                1
                            })
                            .sum::<usize>()
                            == 0
                        {
                            return Err(Error::MissingNextRule);
                        }

                        let _ = opcodes.pop();
                    }
                    ExprOp::Add | ExprOp::Mul | ExprOp::Div | ExprOp::Sub => {
                        let mut iter = compiled_terms.into_iter();
                        let (mut term_opcodes, return_value) =
                            iter.next().ok_or(Error::MissingNextRule)?;
                        opcodes.append(&mut term_opcodes);

                        for (mut term_opcodes, term_return) in iter {
                            opcodes.append(&mut term_opcodes);
                            opcodes.push(match op {
                                ExprOp::Add => OpCode::ADD(return_value, return_value, term_return),
                                ExprOp::Mul => OpCode::MUL(return_value, return_value, term_return),
                                ExprOp::Div => OpCode::DIV(return_value, return_value, term_return),
                                ExprOp::Sub => OpCode::SUB(return_value, return_value, term_return),
                                _ => unreachable!(),
                            });
                        }

                        return Ok((opcodes, return_value));
                    }
                    ExprOp::Or | ExprOp::And => {
                        opcodes = vec![OpCode::LOAD(return_value, Value::Bool(false))];
                        for (mut term_opcodes, term_return_value) in compiled_terms.into_iter() {
                            opcodes.append(&mut term_opcodes);
                            opcodes.push(OpCode::MOV(return_value, term_return_value));
                            match op {
                                ExprOp::Or => opcodes.push(OpCode::JEQ(return_value, exit_label)),
                                ExprOp::And => opcodes.push(OpCode::JNE(return_value, exit_label)),
                                _ => unreachable!(),
                            };
                        }
                    }
                    x => panic!("unimplemented {:#?}", x),
                }

                opcodes.push(OpCode::LABEL(current_exit_label));
                (opcodes, return_value)
            }
        })
    }

    pub fn optimize(opcodes: Vec<OpCode>) -> (Vec<OpCode>, bool) {
        vec![
            optimizations::assign_unique_register_addresses,
            optimizations::remove_redundant_load_and_mov,
            optimizations::remove_unused_values,
            optimizations::calculate_static_values,
            optimizations::remove_useless_jumps,
            optimizations::remove_dead_code,
            //optimizations::move_load_top,
        ]
        .into_iter()
        .fold((opcodes, false), |(opcodes, has_optimized), f| {
            let (opcodes, optimized) = f(opcodes);
            (opcodes, has_optimized || optimized)
        })
    }

    /// Returns a list of Addresses that are labeled in the opcodes list. This function is required
    /// to resolve any jump in the compiler
    pub fn labels_to_addr(opcodes: &[OpCode]) -> HashMap<Addr, (Addr, Addr)> {
        let mut pos_without_labels = 0;
        opcodes
            .iter()
            .enumerate()
            .map(|(actual_pos, opcode)| match opcode {
                OpCode::LABEL(label) => (opcode, pos_without_labels, actual_pos),
                _ => {
                    pos_without_labels = pos_without_labels + 1usize;
                    (opcode, pos_without_labels - 1, actual_pos)
                }
            })
            .filter_map(|(opcode, pos_without_label, pos)| match opcode {
                OpCode::LABEL(id) => Some((*id, (pos_without_label.into(), pos.into()))),
                _ => None,
            })
            .collect::<HashMap<_, _>>()
    }

    pub fn resolve_label_to_addr(opcodes: Vec<OpCode>) -> Result<Vec<OpCode>, Error> {
        let labels_to_addr = Self::labels_to_addr(&opcodes);

        opcodes
            .into_iter()
            .filter(|opcode| !matches!(opcode, OpCode::LABEL(_)))
            .map(|opcode| {
                // Rewrite JMP to not use labels but instead addresses
                Ok(match opcode {
                    OpCode::JMP(label) => OpCode::JMP(
                        labels_to_addr
                            .get(&label)
                            .map(|addr| addr.0)
                            .ok_or(Error::UnknownLabel(*label))?,
                    ),
                    OpCode::JEQ(register, label) => OpCode::JEQ(
                        register,
                        labels_to_addr
                            .get(&label)
                            .map(|addr| addr.0)
                            .ok_or(Error::UnknownLabel(*label))?,
                    ),
                    OpCode::JNE(register, label) => OpCode::JNE(
                        register,
                        labels_to_addr
                            .get(&label)
                            .map(|addr| addr.0)
                            .ok_or(Error::UnknownLabel(*label))?,
                    ),
                    opcode => opcode,
                })
            })
            .collect()
    }

    pub fn compile(mut self) -> Result<Vec<OpCode>, Error> {
        let program_exit = self.next_label();
        let (mut opcodes, return_value) = self.compile_expr(self.expr, program_exit)?;
        opcodes.push(OpCode::LABEL(program_exit));
        opcodes.push(OpCode::HLT(return_value));
        Ok(opcodes)
    }
}
