use crate::filter_expr::{opcode::OpCode, Register};
use std::collections::HashMap;

/// When code is optimized, it is possible that a few registers are being LOADED and not read
/// and it is not caught by `remove_unused_values`. This function will make sure that every
/// LOADED and LOADED_EXTERNAL register has a unique address.
pub fn assign_unique_register_addresses(mut opcodes: Vec<OpCode>) -> (Vec<OpCode>, bool) {
    let mut has_changed = false;
    let mut counter: usize = 0;
    let mut register_new_mapping = HashMap::<_, Register>::new();
    let mut new_mapping =
        |register_new_mapping: &mut HashMap<Register, Register>, current: &Register| -> Register {
            counter += 1;
            if counter != **current {
                has_changed = true;
            }

            register_new_mapping.insert(current.clone(), counter.into());
            counter.into()
        };

    opcodes.iter_mut().for_each(|opcode| match opcode {
        OpCode::LOAD(current, _) | OpCode::LOAD_EXTERNAL(current, _) => {
            *current = new_mapping(&mut register_new_mapping, current);
        }
        OpCode::SHL(set, reg1, _)
        | OpCode::SHR(set, reg1, _)
        | OpCode::NOT(set, reg1)
        | OpCode::CPY(set, reg1)
        | OpCode::NEG(set, reg1)
        | OpCode::MOV(set, reg1) => {
            *reg1 = register_new_mapping[reg1].clone();

            if let Some(new_set) = register_new_mapping.get(set) {
                *set = new_set.clone()
            } else {
                *set = new_mapping(&mut register_new_mapping, set);
            }
        }
        OpCode::HLT(reg1) | OpCode::JEQ(reg1, _) | OpCode::JNE(reg1, _) => {
            *reg1 = register_new_mapping[reg1].clone();
        }
        OpCode::ADD(set, reg1, reg2)
        | OpCode::MUL(set, reg1, reg2)
        | OpCode::NE(set, reg1, reg2)
        | OpCode::GE(set, reg1, reg2)
        | OpCode::DIV(set, reg1, reg2)
        | OpCode::LE(set, reg1, reg2)
        | OpCode::GT(set, reg1, reg2)
        | OpCode::LT(set, reg1, reg2)
        | OpCode::DIV(set, reg1, reg2)
        | OpCode::MOD(set, reg1, reg2)
        | OpCode::XOR(set, reg1, reg2)
        | OpCode::SUB(set, reg1, reg2)
        | OpCode::AND(set, reg1, reg2)
        | OpCode::EQ(set, reg1, reg2)
        | OpCode::OR(set, reg1, reg2) => {
            *reg1 = register_new_mapping[reg1].clone();
            *reg2 = register_new_mapping[reg2].clone();
            if let Some(new_set) = register_new_mapping.get(set) {
                *set = new_set.clone()
            } else {
                *set = new_mapping(&mut register_new_mapping, set);
            }
        }
        OpCode::LABEL(_) | OpCode::JMP(_) => {}
    });

    (opcodes, has_changed)
}
