use crate::filter_expr::opcode::OpCode;
use std::collections::HashMap;

/// Move the LOAD to the Top of the program
///
/// This optimization will allow the VM to execute all the LOAD operations at the beginning of the
/// program, once, then clone the initial state of the registers and execute the rest of them.
///
/// This optimization works under the assumption that cloning the registers is faster than executing
/// it
pub fn move_load_top(mut opcodes: Vec<OpCode>) -> (Vec<OpCode>, bool) {
    let mut load_new_indexes = HashMap::new();
    let mut new_index = opcodes.len();

    if opcodes
        .iter()
        .fold((true, false), |(is_sorted, non_load_stage), opcode| {
            let is_load_opcode = match opcode {
                OpCode::LOAD(reg, _) => {
                    load_new_indexes.insert(*reg, new_index.into());
                    new_index += 1;
                    true
                }
                _ => false,
            };

            if non_load_stage && is_load_opcode {
                (false, true)
            } else if !is_load_opcode {
                (is_sorted, true)
            } else {
                (is_sorted, non_load_stage)
            }
        })
        .0
    {
        // already sorted, all the LOAD opcodes are at the beginning of the program
        return (opcodes, false);
    }

    /// assign new indexes to the static registers and rewrite the entire opcodes to use the new
    /// addresses
    opcodes.iter_mut().for_each(|opcode| match opcode {
        OpCode::LOAD(reg, _) => {
            *reg = load_new_indexes[reg];
        }
        OpCode::LABEL(_) | OpCode::JMP(_) => {}
        OpCode::LOAD_EXTERNAL(dst, _)
        | OpCode::HLT(dst)
        | OpCode::JEQ(dst, _)
        | OpCode::JNE(dst, _) => {
            if let Some(new_dst) = load_new_indexes.get(dst) {
                *dst = *new_dst;
            }
        }
        OpCode::MOV(dst, reg1)
        | OpCode::SHL(dst, reg1, _)
        | OpCode::CPY(dst, reg1)
        | OpCode::SHR(dst, reg1, _)
        | OpCode::NEG(dst, reg1)
        | OpCode::NOT(dst, reg1) => {
            if let Some(value) = load_new_indexes.get(reg1) {
                *reg1 = *value;
            }
            if let Some(new_dst) = load_new_indexes.get(dst) {
                *dst = *new_dst;
            }
        }
        OpCode::EQ(dst, reg1, reg2)
        | OpCode::GE(dst, reg1, reg2)
        | OpCode::GT(dst, reg1, reg2)
        | OpCode::LE(dst, reg1, reg2)
        | OpCode::LT(dst, reg1, reg2)
        | OpCode::NE(dst, reg1, reg2)
        | OpCode::AND(dst, reg1, reg2)
        | OpCode::OR(dst, reg1, reg2)
        | OpCode::XOR(dst, reg1, reg2)
        | OpCode::ADD(dst, reg1, reg2)
        | OpCode::SUB(dst, reg1, reg2)
        | OpCode::MUL(dst, reg1, reg2)
        | OpCode::DIV(dst, reg1, reg2)
        | OpCode::MOD(dst, reg1, reg2) => {
            if let Some(value) = load_new_indexes.get(reg1) {
                *reg1 = *value;
            }
            if let Some(value) = load_new_indexes.get(reg2) {
                *reg2 = *value;
            }
            if let Some(new_dst) = load_new_indexes.get(dst) {
                *dst = *new_dst;
            }
        }
    });

    // split the opcodes into two parts, the LOAD opcodes and the rest
    let mut parts: (Vec<_>, Vec<_>) = opcodes
        .into_iter()
        .partition(|opcode| matches!(opcode, OpCode::LOAD(_, _)));

    // merge both parts
    let mut new_opcodes = parts.0;
    new_opcodes.extend(parts.1);

    (new_opcodes, true)
}
