use crate::filter_expr::{compiler::Compiler, opcode::OpCode};
use std::collections::HashSet;

/// Remove OpCodes that are not executed.
///
/// This function simulates the Jumps the code may have and attempts to execute all possible
/// branches, regardless of the actual values, keeping track of executed opcodes.
///
/// Non executed opcodes are then removed from the list of opcodes.
#[inline]
pub fn remove_dead_code(mut opcodes: Vec<OpCode>) -> (Vec<OpCode>, bool) {
    let total = opcodes.len();
    // list of opcodes that have been executed
    let mut executed_opcodes = vec![0; total];

    // current opcode position
    let mut current = 0;

    let mut future_jumps = HashSet::new();
    // to avoid infinite loop, since the jumps are unconditional, we need to keep track the jumps
    let mut jumped_at = HashSet::new();

    // convert labels to address for the current list of opcodes
    let label_to_addr = Compiler::labels_to_addr(&opcodes);

    loop {
        match if let Some(opcode) = opcodes.get(current) {
            executed_opcodes[current] += 1;
            opcode
        } else {
            break;
        } {
            OpCode::HLT(_) => {
                let mut has_jumped = false;
                for jump_to in future_jumps.iter() {
                    if !jumped_at.contains(jump_to) {
                        jumped_at.insert(*jump_to);
                        has_jumped = true;
                        current = *jump_to;
                        break;
                    }
                }

                if !has_jumped {
                    break;
                }
            }
            OpCode::JMP(to) => {
                let to = label_to_addr
                    .get(to)
                    .map(|addr| addr.1)
                    .expect("Invalid label");
                if !jumped_at.contains(&to) {
                    jumped_at.insert(*to);
                    current = *to;
                } else {
                    current += 1;
                }
            }
            OpCode::JEQ(_, to) | OpCode::JNE(_, to) => {
                let to = label_to_addr
                    .get(to)
                    .map(|addr| addr.1)
                    .expect("Invalid label");
                future_jumps.insert(*to);
                current += 1;
            }
            _ => {
                current += 1;
            }
        }
    }

    let total_opcodes = opcodes.len();
    let new_opcodes = opcodes
        .into_iter()
        .enumerate()
        .filter_map(|(idx, opcode)| {
            if executed_opcodes[idx] == 0 {
                None
            } else {
                Some(opcode)
            }
        })
        .collect::<Vec<_>>();

    let new_opcodes_total = new_opcodes.len();

    (new_opcodes, total_opcodes != new_opcodes_total)
}
