use crate::filter_expr::opcode::OpCode;
use std::collections::HashMap;

/// Remove loaded values that are not read by any opcode. This is useful for reducing the size
/// of the opcodes to be executed
pub fn remove_unused_values(mut opcodes: Vec<OpCode>) -> (Vec<OpCode>, bool) {
    let used_registers = opcodes
        .iter()
        .enumerate()
        .map(|(pos, opcode)| match opcode {
            OpCode::LOAD(_, _)
            | OpCode::LOAD_EXTERNAL(_, _)
            | OpCode::LABEL(_)
            | OpCode::JMP(_) => vec![],
            OpCode::NOT(_, reg1)
            | OpCode::CPY(_, reg1)
            | OpCode::MOV(_, reg1)
            | OpCode::JEQ(reg1, _)
            | OpCode::SHL(_, reg1, _)
            | OpCode::SHR(_, reg1, _)
            | OpCode::NEG(_, reg1)
            | OpCode::HLT(reg1)
            | OpCode::JNE(reg1, _)
            | OpCode::JEQ(reg1, _) => {
                vec![(*reg1, pos)]
            }
            OpCode::ADD(_, reg1, reg2)
            | OpCode::GE(_, reg1, reg2)
            | OpCode::LE(_, reg1, reg2)
            | OpCode::MUL(_, reg1, reg2)
            | OpCode::NE(_, reg1, reg2)
            | OpCode::DIV(_, reg1, reg2)
            | OpCode::MOD(_, reg1, reg2)
            | OpCode::XOR(_, reg1, reg2)
            | OpCode::SUB(_, reg1, reg2)
            | OpCode::AND(_, reg1, reg2)
            | OpCode::OR(_, reg1, reg2)
            | OpCode::EQ(_, reg1, reg2)
            | OpCode::GT(_, reg1, reg2)
            | OpCode::LT(_, reg1, reg2) => {
                vec![(*reg1, pos), (*reg2, pos)]
            }
        })
        .flatten()
        .fold(HashMap::new(), |mut acc, (reg, pos)| {
            acc.entry(reg).or_insert_with(Vec::new).push(pos);
            acc
        });

    let old_total_opcodes = opcodes.len();

    // remove unused registers. If the register has not been read by any opcode, then it can be
    // removed.
    let new_opcodes = opcodes
        .into_iter()
        .enumerate()
        .filter_map(|(pos, opcode)| match &opcode {
            OpCode::LOAD(dst, _) | OpCode::LOAD_EXTERNAL(dst, _) => {
                if used_registers.get(dst).is_some() {
                    return Some(opcode);
                } else {
                    None
                }
            }
            _ => Some(opcode),
        })
        .collect::<Vec<_>>();

    let new_total_opcodes = new_opcodes.len();
    (new_opcodes, old_total_opcodes != new_total_opcodes)
}
